from cython cimport floating
cimport numpy as cnp

# --------------------------------------------------
# Define fused data types:

ctypedef fused np_floating:
    cnp.float32_t
    cnp.float64_t

ctypedef fused int_dtype:
    cnp.int32_t
    cnp.int64_t

ctypedef fused noncomplex_numeric:
    cnp.int8_t
    cnp.int16_t
    cnp.int32_t
    cnp.int64_t
    cnp.float32_t
    cnp.float64_t

# --------------------------------------------------

cdef void cpp_blas_axpy(floating[::1] v1, floating[::1] v2, floating alpha) noexcept nogil
cdef floating cpp_blas_dot(floating[::1] v1, floating[::1] v2) noexcept nogil

cpdef floating cpp_update_delta(floating n,
                        floating l0,
                        floating l1,
                        floating theta,
                        floating var) noexcept nogil

cpdef void cpp_update_delta_alpha_vec(
                        floating[::1] delta,   
                        floating n,            
                        floating theta,      
                        floating var,          
                        floating[::1] l0_vec,  
                        floating[::1] l1_vec,  
                        int threads) noexcept nogil

cpdef floating cpp_update_var(floating n,
                        floating[::1] beta,
                        floating[::1] std_beta,
                        floating[::1] q,
                        floating lam_min) noexcept nogil

cpdef void cpp_update_beta_ssl(int[::1] ld_left_bound,
                        int_dtype[::1] ld_indptr,
                        noncomplex_numeric[::1] ld_data,
                        floating[::1] std_beta,
                        floating[::1] beta,
                        floating[::1] beta_diff,
                        floating[::1] q,
                        floating[::1] n_per_snp,
                        int[::1] p_gamma,
                        floating[::1] theta,
                        floating[::1] delta,
                        floating[::1] var,
                        floating n,
                        floating l0,
                        floating l1,
                        floating a,
                        floating b,
                        int update_freq,
                        floating lam_min,
                        floating init_var,
                        floating min_var,
                        bint u_var,
                        floating dq_scale,
                        int threads,
                        bint low_memory) noexcept nogil

cpdef void cpp_update_beta_ssl_alpha(int[::1] ld_left_bound,
                        int_dtype[::1] ld_indptr,
                        noncomplex_numeric[::1] ld_data,
                        floating[::1] std_beta,
                        floating[::1] beta,
                        floating[::1] beta_diff,
                        floating[::1] q,
                        floating[::1] n_per_snp,
                        int[::1] p_gamma,
                        floating[::1] theta,
                        floating[::1] delta,
                        floating[::1] var,
                        floating n,
                        floating[::1] l0_vec,
                        floating[::1] l1_vec,
                        floating a,
                        floating b,
                        int update_freq,
                        floating lam_min,
                        floating init_var,
                        floating min_var,
                        bint u_var,
                        floating dq_scale,
                        int threads,
                        bint low_memory) noexcept nogil


cpdef void cpp_update_beta_lasso(int[::1] ld_left_bound,
                        int_dtype[::1] ld_indptr,
                        noncomplex_numeric[::1] ld_data,
                        floating[::1] std_beta,
                        floating[::1] beta,
                        floating[::1] beta_diff,
                        floating[::1] q,
                        floating[::1] n_per_snp,
                        floating lam,
                        floating lam_min,
                        floating dq_scale,
                        int threads,
                        bint low_memory) noexcept nogil



import pandas as pd
import numpy as np
from multiprocessing import shared_memory
from pprint import pprint
from joblib import Parallel, delayed
from loky import get_reusable_executor
import copy
from tqdm import tqdm
import time

from sslprs.model.SSL import SSL

def fit_model_fixed_params(model, fixed_params, scale_l0_by_l1=False, shm_data=None, **fit_kwargs):
    """

    Perform model fitting using a set of fixed set of (hyper)-parameters.
    This is a helper function to allow users to use the `multiprocessing` module
    to fit PRS models in parallel.

    :param model: A PRS model object that implements a `.fit()` method and takes `fix_params` as an attribute.
    :param fixed_params: A dictionary of fixed parameters to use for the model fitting.
    :param shm_data: A dictionary of shared memory data to use for the model fitting. This is primarily used to
    share LD data across multiple processes.
    :param fit_kwargs: Key-word arguments to pass to the `.fit()` method of the PRS model.

    :return: The fitted PRS model. If the fitting fails, the function returns `None`.
    """
    
    model.fix_params = fixed_params

    if scale_l0_by_l1:
        model.fix_params["l0"] = model.fix_params["l0"]*model.fix_params["l1"]
        # print(model.fix_params["l0"])

    if shm_data is not None:

        model.ld_data = {}

        try:
            ld_data_shm = shared_memory.SharedMemory(name=shm_data['shm_name'])
            model.ld_data[shm_data['chromosome']] = np.ndarray(
                shape=shm_data['shm_shape'],
                dtype=shm_data['shm_dtype'],
                buffer=ld_data_shm.buf
            )
        except FileNotFoundError:
            raise Exception("LD data not found in shared memory.")

    try:
        model.fit(**fit_kwargs)
    except Exception as e:
        model = None
    finally:
        if shm_data is not None:
            ld_data_shm.close()
            model.ld_data = None

    if model is not None:
        return {
            'coef_table': model.to_table()[['CHR', 'SNP', 'POS', 'A1', 'A2', 'BETA']],
            'hyp_table': model.to_theta_table(),
            'training_objective': model.obj_f()
        }

def combine_coefficient_tables(coef_tables, coef_col='BETA'):
    """
    Combine a list of coefficient tables (output from a PRS model) into a single
    table that can be used for downstream tasks, such scoring and evaluation. Note that
    this implementation assumes that the coefficients tables were generated for the same
    set of variants, from a grid-search or similar procedure.

    :param coef_tables: A list of pandas dataframes containing variant information as well as
    inferred coefficients.
    :param coef_col: The name of the column containing the coefficients.
    :return: A single pandas dataframe with the combined coefficients. The new coefficient columns will be
    labelled as BETA_0, BETA_1, etc.
    """

    # Sanity checks:
    assert all([coef_col in t.columns for t in coef_tables]), "All tables must contain the coefficient column."
    assert all([len(t) == len(coef_tables[0]) for t in coef_tables]), "All tables must have the same number of rows."

    if len(coef_tables) == 1:
        return coef_tables[0]

    ref_table = coef_tables[0].copy()
    ref_table.rename(columns={coef_col: f'{coef_col}_0'}, inplace=True)

    # Extract the coefficients from the other tables:
    return pd.concat([ref_table, *[t[[coef_col]].rename(columns={coef_col: f'{coef_col}_{i}'})
                                   for i, t in enumerate(coef_tables[1:], 1)]], axis=1)

class HyperparameterSearch(object):

    def __init__(self,
                 gdl,
                 model=None,
                 criterion = "pseudo_validation",
                 validation_gdl=None,
                 verbose=False,
                 n_jobs=1,
                 threads=1):

        self.gdl = gdl
        self.n_jobs = n_jobs
        self.threads=threads

        # print(self.threads)
        if model is None:
            self.model = SSL(gdl, threads = self.threads)
        else:
            import inspect
            if inspect.isclass(model):
                self.model = model(gdl, threads = self.threads)
            else:
                self.model = model

        self.validation_result = None

        self.criterion = criterion
        self._validation_gdl = validation_gdl

        self.verbose = verbose
        self.model.verbose = verbose

        if self._validation_gdl is not None:
            self._validation_gdl.verbose = verbose

        if self.criterion == 'pseudo_validation':
            assert self._validation_gdl is not None
            assert self._validation_gdl.sumstats_table is not None

    def to_validation_table(self):
        """
        Summarize the validation results in a pandas table.
        """
        if self.validation_result is None:
            raise Exception("Validation result is not set!")
        elif len(self.validation_result) < 1:
            raise Exception("Validation result is not set!")

        return pd.DataFrame(self.validation_result)

    def write_validation_result(self, v_filename, sep="\t"):
        """
        After performing hyperparameter search, write a table
        that records that value of the objective for each combination
        of hyperparameters.
        :param v_filename: The filename for the validation table.
        :param sep: The separator for the validation table
        """

        v_df = self.to_validation_table()
        v_df.to_csv(v_filename, index=False, sep=sep)

    def multi_objective(self, models):
        """
        This method evaluates multiple PRS models simultaneously. This can be faster for
        some evaluation criteria, such as the validation R^2, because we only need to
        multiply the inferred effect sizes with the genotype matrix only once.

        :param models: A list of PRS models that we wish to evaluate.
        """

        if len(models) == 1:
            return self.objective(models[0])

        elif self.criterion == 'pseudo_validation':
            return [m.pseudo_validate(self._validation_gdl) for m in models]

    def objective(self, model):
        """
        A method that takes the result of fitting the model
        and returns the desired objective (either `ELBO`, `pseudo_validation`, or `validation`).
        :param model: The PRS model to evaluate
        """

        if self.criterion == 'pseudo_validation':
            return model.pseudo_validate(self._validation_gdl)
    def _evaluate_models(self):
        """
        This method evaluates multiple PRS models to determine their relative performance based on the
        criterion set by the user. The criterion can be the training objective (e.g. ELBO in the case of VIPRS),
        pseudo-validation or validation using held-out test data.

        :return: The metrics associated with each model setup.
        """

        assert self._training_objective is not None
        assert self._model_coefs is not None

        if self.criterion == 'training_objective':
            metrics = self._training_objective
        elif self.criterion == 'pseudo_validation':
            from penprs.eval.pseudo_metrics import pseudo_r2

            metrics = pseudo_r2(self._validation_gdl, self._model_coefs)

        return metrics
    
    def fit(self):
        raise NotImplementedError
    
class GridSearch(HyperparameterSearch):
    """
    Hyperparameter search using Grid Search
    """

    def __init__(self,
                 gdl,
                 grid,
                 model=None,
                 criterion='pseudo_validation',
                 validation_gdl=None,
                 verbose=False,
                 scale_l0_by_l1 = False,
                 n_jobs=1,
                 threads=1):

        """
        Perform hyperparameter search using grid search
        :param gdl: A GWADataLoader object
        :param grid: A HyperParameterGrid object
        :param model: A `PRSModel`-derived object (e.g. VIPRS).
        :param criterion: The objective function for the grid search (ELBO or validation).
        :param validation_gdl: If the objective is validation, provide the GWADataLoader object for the validation
        dataset.
        :param verbose: Detailed messages and print statements.
        :param n_jobs: The number of processes to use for the grid search
        """

        super().__init__(gdl, model=model, criterion=criterion,
                         validation_gdl=validation_gdl,
                         verbose=verbose,
                         n_jobs=n_jobs,
                         threads=threads)

        self.grid = grid
        self.scale_l0_by_l1 = scale_l0_by_l1

    def fit(self, max_iter=1000, abs_tol=1e-6):

        """
        Perform grid search over the hyperparameters to determine the
        best model based on the criterion set by the user. This utility method
        performs model fitting across the grid of hyperparameters, potentially in parallel
        if `n_jobs` is greater than 1.

        :param max_iter: The maximum number of iterations to run for each model fit.
        :param f_abs_tol: The absolute tolerance for the function convergence criterion.
        :param x_abs_tol: The absolute tolerance for the parameter convergence criterion.

        :return: The best model based on the criterion set by the user.
        """

        #keep track of fitting and validation time
        fit_valid_time = {}

        fit_start_time = time.time()

        print("> Performing Grid Search over the following grid:")
        print(self.grid.to_table())

        if self.n_jobs > 1:
            # Only create the shared memory object if the number of processes is more than 1.
            # Otherwise, this would be a waste of resources.

            # ----------------- Copy the LD data to shared memory -----------------
            ld_data_arr = self.model.ld_data
            # Create a shared memory block for the array
            shm = shared_memory.SharedMemory(create=True, size=ld_data_arr.nbytes)

            # Create a NumPy array backed by the shared memory block
            shared_array = np.ndarray(ld_data_arr.shape, dtype=ld_data_arr.dtype, buffer=shm.buf)

            np.copyto(shared_array, ld_data_arr)

            del ld_data_arr
            self.model.ld_data = None

            shm_args = {
                'shm_name': shm.name,
                'chromosome': self.model.chr,
                'shm_shape': shared_array.shape,
                'shm_dtype': shared_array.dtype
            }

        else:
            shm_args = None

        # --------------------------------------------------------------------
        # Perform grid search:

        grid = self.grid.combine_grids()

        parallel = Parallel(n_jobs=self.n_jobs, backend="multiprocessing")

        with parallel:

            fitted_models = parallel(
                delayed(fit_model_fixed_params)(self.model, g, self.scale_l0_by_l1, shm_args,
                                                max_iter=max_iter,
                                                abs_tol=abs_tol)
                for g in tqdm(grid, desc="Fitting Models")
            )

        # Clean up after performing model fit:
        self.model.ld_data = None  # To minimize memory usage with validation/pseudo-validation

        # Close and unlink shared memory objects:
        if shm_args is not None:
            shm.close()
            shm.unlink()

        fit_end_time = time.time()

        fit_valid_time["Fit_time"] = round(fit_end_time - fit_start_time, 2)

        # --------------------------------------------------------------------
        # Post-process the results and determine the best model:

        valid_start_time = time.time()

        assert not all([fm is None for fm in fitted_models]), "None of the models converged successfully."

        self._model_coefs = combine_coefficient_tables([fm['coef_table'] for fm in fitted_models if fm is not None])
        self._model_hyperparams = [fm['hyp_table'] for fm in fitted_models if fm is not None]
        self._training_objective = [fm['training_objective'] for fm in fitted_models if fm is not None]

        # 2) Perform evaluation on the models that converged:
        eval_metrics = self._evaluate_models()

        # 3) Combine all the results together into a single table (populate records in
        # self.validation_result):

        self.validation_result = []
        success_counter = 0

        for i, vr in enumerate(grid):
            if fitted_models[i] is not None:
                vr['Converged'] = True
                vr['training_objective'] = self._training_objective[success_counter]
                if self.criterion != 'training_objective':
                    vr[self.criterion] = eval_metrics[success_counter]
                success_counter += 1
            else:
                vr['Converged'] = False
                vr['training_objective'] = np.NaN
                if self.criterion != 'training_objective':
                    vr[self.criterion] = np.NaN

            self.validation_result.append(vr)

        # --------------------------------------------------------------------
        # Determine and return the best model:
        validation_table = self.to_validation_table()
        best_idx = np.argmax(validation_table[self.criterion].values)

        print("> Grid search identified the best hyperparameters as:")
        pprint(grid[best_idx])

        self.model.fix_params = grid[best_idx]
        self.model.initialize()
        self.model.betas=fitted_models[best_idx]['coef_table']["BETA"]
        # self.model.set_model_parameters(fitted_models[best_idx]['coef_table'])

        valid_end_time = time.time()

        fit_valid_time["Validation_time"] = valid_end_time - valid_start_time

        return self.model, fit_valid_time, validation_table

    # def select_best_model(self):

    #     if len(self.fit_results) > 1:
    #         res_objectives = self.multi_objective(self.fit_results)
    #     else:
    #         raise Exception("Error: Convergence was achieved for less than 2 models.")

    #     if self.criterion == 'pseudo_validation':
    #         for i in range(len(self.validation_result)):
    #             self.validation_result[i]['Pseudo_Validation_Corr'] = res_objectives[i]

    #     if self.criterion == "pseudo_validation":
    #         best_idx = np.argmax(res_objectives)
    #         best_params = self.params[best_idx]

    #     print("> Grid search identified the best hyperparameters as:")
    #     pprint(best_params)

    #     print("> Refitting the model with the best hyperparameters...")

    #     self.model.fix_params = best_params
    #     self.model.fit()

    #     return self
    def to_table(self):
        return self.model.to_table()


import numpy as np
from viprs.eval.pseudo_metrics import _match_variant_stats

def pseudo_r2(test_gdl, prs_beta_table):
    """
    `pseudo_r2` metric as in the viprs package: https://github.com/shz9/viprs
    Provides better validation curves and variable selection when used to 
    evaluate R-Squared in grid search than pearson_pseudo_r
    
    Compute the R-Squared metric (proportion of variance explained) for a given
    PRS using standardized marginal betas from an independent test set.
    Here, we follow the pseudo-validation procedures outlined in Mak et al. (2017) and
    Yang and Zhou (2020), where the proportion of phenotypic variance explained by the PRS
    in an independent validation cohort can be approximated with:

    R2(PRS, y) ~= 2*r'b - b'Sb

    Where `r` is the standardized marginal beta from a validation/test set,
    `b` is the posterior mean for the effect size of each variant and `S` is the LD matrix.

    :param test_gdl: An instance of `GWADataLoader` with the summary statistics table initialized.
    :param prs_beta_table: A pandas DataFrame with the PRS effect sizes. Must contain
    the columns: CHR, SNP, A1, A2, BETA.
    """

    std_beta, prs_beta, q = _match_variant_stats(test_gdl, prs_beta_table)

    rb = np.sum((prs_beta.T * std_beta).T, axis=0)
    bsb = np.sum(prs_beta*q, axis=0)

    return 2*rb - bsb


def pseudo_pearson_r2(test_gdl, prs_beta_table):
    """
    `pseudo_pearson_r2` metric as in the viprs package: https://github.com/shz9/viprs
    Used during results evaluation as pseudo_r2 above can be sensitive to sparsified LD
    especially on external methods.

    Perform pseudo-validation of the inferred effect sizes by comparing them to
    standardized marginal betas from an independent validation set. Here, we follow the pseudo-validation
    procedures outlined in Mak et al. (2017) and Yang and Zhou (2020), where
    the correlation between the PRS and the phenotype in an independent validation
    cohort can be approximated with:

    Corr(PRS, y) ~= r'b / sqrt(b'Sb)

    Where `r` is the standardized marginal beta from a validation set,
    `b` is the posterior mean for the effect size of each variant and `S` is the LD matrix.

    :param test_gdl: An instance of `GWADataLoader` with the summary statistics table initialized.
    :param prs_beta_table: A pandas DataFrame with the PRS effect sizes. Must contain
    the columns: CHR, SNP, A1, A2, BETA.
    """

    std_beta, prs_beta, q = _match_variant_stats(test_gdl, prs_beta_table)

    rb = np.sum((prs_beta.T * std_beta).T, axis=0)
    bsb = np.sum(prs_beta * q, axis=0)

    return (rb / np.sqrt(bsb))**2


# distutils: language = c++

from cython cimport floating

# ------------------------------------------
cdef extern from "co_utils.hpp" nogil:

    bint blas_supported() noexcept nogil
    bint omp_supported() noexcept nogil

    void blas_axpy[T](T* y, T* x, T alpha, int size) noexcept nogil
    T blas_dot[T](T* x, T* y, int size) noexcept nogil

cdef extern from "coord_ascent_ssl.hpp" nogil:

    T update_delta[T](T n,
             T l0,
             T l1,
             T theta,
             T var) noexcept nogil
    
    T update_var[T](int c_size,
            T n,
            T *beta,
            T *std_beta,
            T *q,
            T lam_min) noexcept nogil

    void update_beta_ssl[T, U, I](int c_size,
            T n,
            T l0,
            T l1,
            T a,
            T b,
            int update_freq,
            T lam_min,
            int *ld_left_bound,
            I *ld_indptr,
            U *ld_data,
            T *std_beta,
            T *beta,
            T *beta_diff,
            T *theta,
            T *delta,
            T *var,
            T *q,
            T *n_per_snp,
            T init_var,
            T min_var,
            bint u_var,
            T dq_scale,
            int *p_gamma,
            int threads,
            bint low_memory) noexcept nogil

cdef extern from "coord_ascent_ssl_alpha.hpp" nogil:
    void update_delta_vec[T](int c_size,
                            T n,
                            T theta,
                            T var,
                            T* l0_vec,
                            T* l1_vec,
                            T* delta,
                            int threads) noexcept nogil

    void update_beta_ssl_alpha[T, U, I](int c_size,
            T n,
            T *l0_vec,
            T *l1_vec,
            T a,
            T b,
            int update_freq,
            T lam_min,
            int *ld_left_bound,
            I *ld_indptr,
            U *ld_data,
            T *std_beta,
            T *beta,
            T *beta_diff,
            T *theta,
            T *delta,
            T *var,
            T *q,
            T *n_per_snp,
            T init_var,
            T min_var,
            bint u_var,
            T dq_scale,
            int *p_gamma,
            int threads,
            bint low_memory) noexcept nogil

cdef extern from "coord_ascent_lasso.hpp" nogil:

    void update_beta_lasso[T, U, I](int c_size,
            int *ld_left_bound,
            I *ld_indptr,
            U *ld_data,
            T *std_beta,
            T *beta,
            T *beta_diff,
            T *q,
            T *n_per_snp,
            T lam,
            T lam_min,
            T dq_scale,
            int threads,
            bint low_memory) noexcept nogil

# ------------------------------------------

cpdef check_blas_support():
    return blas_supported()

cpdef check_omp_support():
    return omp_supported()

cdef void cpp_blas_axpy(floating[::1] v1, floating[::1] v2, floating alpha) noexcept nogil:
    """v1 := v1 + alpha * v2"""
    cdef int size = v1.shape[0]
    blas_axpy(&v1[0], &v2[0], alpha, size)

cdef floating cpp_blas_dot(floating[::1] v1, floating[::1] v2) noexcept nogil:
    """v1 := v1.Tv2"""
    cdef int size = v1.shape[0]
    return blas_dot(&v1[0], &v2[0], size)


# ---------------- SSL Coordinate Ascent Functions ----------------

cpdef floating cpp_update_delta(floating n,
                        floating l0,
                        floating l1,
                        floating theta,
                        floating var) noexcept nogil:
        return update_delta(n,
                    l0,
                    l1,
                    theta,
                    var)
                    
cpdef void cpp_update_delta_alpha_vec(
                        floating[::1] delta,
                        floating n,
                        floating theta,
                        floating var,
                        floating[::1] l0_vec,
                        floating[::1] l1_vec,
                        int threads) noexcept nogil:
    cdef int c_size = delta.shape[0]
    update_delta_vec(c_size,
                     n,
                     theta,
                     var,
                     &l0_vec[0],
                     &l1_vec[0],
                     &delta[0],
                     threads)


cpdef floating cpp_update_var(floating n,
                        floating[::1] beta,
                        floating[::1] std_beta,
                        floating[::1] q,
                        floating lam_min) noexcept nogil:
        return update_var(beta.shape[0],
                    n,
                    &beta[0],
                    &std_beta[0],
                    &q[0],
                    lam_min)

cpdef void cpp_update_beta_ssl(int[::1] ld_left_bound,
                        int_dtype[::1] ld_indptr,
                        noncomplex_numeric[::1] ld_data,
                        floating[::1] std_beta,
                        floating[::1] beta,
                        floating[::1] beta_diff,
                        floating[::1] q,
                        floating[::1] n_per_snp,
                        int[::1] p_gamma,
                        floating[::1] theta,
                        floating[::1] delta,
                        floating[::1] var,
                        floating n,
                        floating l0,
                        floating l1,
                        floating a,
                        floating b,
                        int update_freq,
                        floating lam_min,
                        floating init_var,
                        floating min_var,
                        bint u_var,
                        floating dq_scale,
                        int threads,
                        bint low_memory) noexcept nogil:

        update_beta_ssl(beta.shape[0],
                    n,
                    l0,
                    l1,
                    a,
                    b,
                    update_freq,
                    lam_min,
                    &ld_left_bound[0],
                    &ld_indptr[0],
                    &ld_data[0],
                    &std_beta[0],
                    &beta[0],
                    &beta_diff[0],
                    &theta[0],
                    &delta[0],
                    &var[0],
                    &q[0],
                    &n_per_snp[0],
                    init_var,
                    min_var,
                    u_var,
                    dq_scale,
                    &p_gamma[0],
                    threads,
                    low_memory)

cpdef void cpp_update_beta_ssl_alpha(int[::1] ld_left_bound,
                        int_dtype[::1] ld_indptr,
                        noncomplex_numeric[::1] ld_data,
                        floating[::1] std_beta,
                        floating[::1] beta,
                        floating[::1] beta_diff,
                        floating[::1] q,
                        floating[::1] n_per_snp,
                        int[::1] p_gamma,
                        floating[::1] theta,
                        floating[::1] delta,
                        floating[::1] var,
                        floating n,
                        floating[::1] l0_vec,
                        floating[::1] l1_vec,
                        floating a,
                        floating b,
                        int update_freq,
                        floating lam_min,
                        floating init_var,
                        floating min_var,
                        bint u_var,
                        floating dq_scale,
                        int threads,
                        bint low_memory) noexcept nogil:

        update_beta_ssl_alpha(beta.shape[0],
                    n,
                    &l0_vec[0],
                    &l1_vec[0],
                    a,
                    b,
                    update_freq,
                    lam_min,
                    &ld_left_bound[0],
                    &ld_indptr[0],
                    &ld_data[0],
                    &std_beta[0],
                    &beta[0],
                    &beta_diff[0],
                    &theta[0],
                    &delta[0],
                    &var[0],
                    &q[0],
                    &n_per_snp[0],
                    init_var,
                    min_var,
                    u_var,
                    dq_scale,
                    &p_gamma[0],
                    threads,
                    low_memory)

# ---------------- Lasso Coordinate Descent functions ----------------


cpdef void cpp_update_beta_lasso(int[::1] ld_left_bound,
                        int_dtype[::1] ld_indptr,
                        noncomplex_numeric[::1] ld_data,
                        floating[::1] std_beta,
                        floating[::1] beta,
                        floating[::1] beta_diff,
                        floating[::1] q,
                        floating[::1] n_per_snp,
                        floating lam,
                        floating lam_min,
                        floating dq_scale,
                        int threads,
                        bint low_memory) noexcept nogil:

        update_beta_lasso(beta.shape[0],
                    &ld_left_bound[0],
                    &ld_indptr[0],
                    &ld_data[0],
                    &std_beta[0],
                    &beta[0],
                    &beta_diff[0],
                    &q[0],
                    &n_per_snp[0],
                    lam,
                    lam_min,
                    dq_scale,
                    threads,
                    low_memory)



import numpy as np
import pandas as pd
from tqdm import tqdm
import logging

from .PRSModel import PRSModel
from ..utils.exceptions import OptimizationDivergence
from .co.coord_opt import cpp_update_beta_lasso


class Lasso(PRSModel):

    def __init__(self,
                 gdl,
                 lam=None,
                 **prs_model_kwargs):

        super().__init__(gdl, **prs_model_kwargs)
        self.lam = lam

    def initialize_hyperparameters(self, theta_0=None):

        if theta_0 is None:
            if self.fix_params is not None:
                theta_0 = self.fix_params
            else:
                theta_0 = {}

        # Default settings for the Lasso hyperparameters:
        lam_max = self._compute_lambda_max()

        if 'lambda' in theta_0:
            self.lam = theta_0['lambda']
        elif 'min_lambda_frac' in theta_0:
            self.lam = theta_0['min_lambda_frac'] * lam_max
        elif self.lam is None:
            # By default, scale lambda_max by 10^-3:
            self.lam = 1e-3 * lam_max

        self.lambda_min = np.dtype(self.float_precision).type(self.lambda_min)
        self.lam = np.dtype(self.float_precision).type(self.lam)

    def _coord_ascent_step(self):
        """
        Perform a single iteration of the coordinate ascent algorithm
        for the Lasso model, given the current settings of the
        coefficients and hyperparameters.
        """

        cpp_update_beta_lasso(
            self.ld_left_bound, self.ld_indptr, self.ld_data,
            self.std_beta,
            self.betas, self.betas_diff,
            self.q, self.n_per_snp,
            self.lam,
            self.lambda_min,
            self.dequantize_scale,
            self.threads,
            self.low_memory
        )

    def penalty(self):
        """
        :return: The Lasso penalty.
        """
        return self.lam * np.sum(np.abs(self.betas), axis=0)

    def objective(self):

        return -.5*self.n*self.mse() - self.penalty()

    def fit(self,
            max_iter=1000,
            theta_0=None,
            param_0=None,
            continued=False,
            min_iter=3,
            f_abs_tol=1e-4,
            x_abs_tol=1e-4,
            disable_pbar=False):

        # initialize parameters
        if not continued:
            self.initialize(theta_0, param_0)
            self.update_history()

        pbar = tqdm(range(max_iter),
                    disable=not self.verbose or disable_pbar,
                    desc=f'Chromosome {self.chromosome} ({self.n_snps} variants)')

        curr_obj = prev_obj = self.history['objective'][-1]
        stop_iter = False

        for i in pbar:

            if stop_iter:
                pbar.set_postfix({
                    'Final objective': f"{curr_obj:.3f}",
                    'Nonzero': self.nnz
                })
                pbar.n = i - 1
                pbar.total = i - 1
                pbar.refresh()
                pbar.close()
                break

            # Run the coordinate ascent step:
            self._coord_ascent_step()
            # Update the tracked parameters (including objectives):
            self.update_history()

            # Extract the current value for the objective:
            curr_obj = self.history['objective'][-1]

            # Update the progress bar:
            pbar.set_postfix(
                {
                    'Objective': f"{curr_obj:.3f}",
                }
            )

            if i > min_iter:
                # Check for convergence:
                if not np.isfinite(curr_obj):
                    raise OptimizationDivergence(f"Stopping at iteration {i}: "
                                                 "The optimization algorithm is not converging!\n"
                                                 f"The objective is undefined ({curr_obj}).")
                elif self.mse() < 0:
                    raise OptimizationDivergence(f"Stopping at iteration {i}: "
                                                 "The optimization algorithm is not converging!\n"
                                                 f"The MSE is negative ({self.mse()}).")
                elif np.isclose(prev_obj, curr_obj, atol=f_abs_tol, rtol=0.):
                    stop_iter = True
                elif self.max_betas_diff < x_abs_tol:
                    stop_iter = True

        return self

    def pathwise_fit(self,
                     lambda_path=None,
                     path_steps=20,
                     max_lambda_frac=.99,
                     largest_first=True,
                     save_intermediate=False,
                     theta_0=None,
                     param_0=None,
                     **fit_kwargs):

        """
        Fit the model using a warm-start strategy, where the model is fit
        multiple times with different values of the l0 hyperparameter.

        :param lambda_path: A list of values for the lambda hyperparameter to use
            in the warm-start strategy. If None, a default ladder of values
            will be used.
        :param path_steps: The number of steps to use in the ladder of lambda values.
        :param max_lambda_frac: The maximum value of the lambda hyperparameter.
        :param largest_first: If True, the ladder of lambda values will be
            sorted in descending order.
        :param save_intermediate: If True, the model parameters will be saved
            for each `l0` value.
        :param theta_0: The initial values for the hyperparameters.
        :param param_0: The initial values for the model parameters.
        :param fit_kwargs: Additional keyword arguments to pass to the `fit`
        """

        assert lambda_path is not None or path_steps is not None, (
            "Either `lambda_path` or `path_steps` must be provided."
        )

        self.initialize(theta_0=theta_0, param_0=param_0)

        if lambda_path is None:
            # Generate a default ladder of lambda values
            # by interpolating from lambda_max to
            # the current lambda value.
            lam_max = max_lambda_frac*self._compute_lambda_max()
            lambda_path = np.logspace(np.log2(self.lam),
                np.log2(lam_max),
                num=path_steps,
                base=2)

        lambda_path = np.sort(lambda_path)
        if largest_first:
            lambda_path = lambda_path[::-1]

        if save_intermediate:
            betas = np.empty((self.n_snps, len(lambda_path)), dtype=self.float_precision)
            qs = np.empty((self.n_snps, len(lambda_path)), dtype=self.float_precision)

        pbar = tqdm(range(len(lambda_path)),
                    total=len(lambda_path),
                    disable=not self.verbose,
                    desc=f'Chromosome {self.chromosome} ({self.n_snps} variants)')

        for i in pbar:

            self.lam = lambda_path[i]
            if i == 0:
                self.update_history()

            self.fit(continued=True, disable_pbar=True, **fit_kwargs)

            pbar.set_postfix(
                {
                    'Objective': f"{self.objective():.3f}",
                    'Nonzero': self.nnz,
                    'Lambda': f"{self.lam:.2f}"
                }
            )

            if save_intermediate:
                betas[:, i] = self.betas
                qs[:, i] = self.q

        if save_intermediate:
            self.betas = betas
            self.q = qs
            self.lam = lambda_path

        return self

    def select_best_model(self, validation_gdl=None, criterion='objective'):

        # Call the parent method:
        best_idx = super().select_best_model(validation_gdl=validation_gdl, criterion=criterion)

        # Update the hyperparameter:
        self.lam = self.lam[best_idx]

    def to_hyperparameter_table(self):
        """
        :return: A pandas DataFrame containing the hyperparameters
            of the model.
        """

        return pd.DataFrame([
            {'Parameter': 'lambda', 'Value': self.lam},
            {'Parameter': 'Nonzero coefficients', 'Value': self.nnz},
            {'Parameter': 'lambda_min', 'Value': self.lambda_min},
        ])



import numpy as np
import pandas as pd
import os.path as osp
import logging
from viprs.utils.compute_utils import expand_column_names


class PRSModel:
    """
    A base class for PRS models.

    This class defines the basic structure and methods
    that are common to most PRS models. Specifically, this class provides methods and interfaces
    for initialization, harmonization, prediction, and fitting of PRS models.

    The class is generic is designed to be inherited and extended by
    a variety of PRS models, such as `Lasso` and `SSL`.

    :ivar gdl: A GWADataLoader object containing harmonized GWAS summary statistics and
    Linkage-Disequilibrium (LD) matrices.
    :ivar Nj: A numpy array containing the sample sizes per variant.
    :ivar shapes: The shapes of the variant arrays (e.g. the number of variants per chromosome).
    :ivar _sample_size: The maximum per-SNP sample size.
    :ivar betas: The coefficients for the variants.
    """

    def __init__(self,
                 gdl,
                 fix_params=None,
                 tracked_params=None,
                 float_precision='float32',
                 lambda_min=None,
                 verbose=True,
                 low_memory=False,
                 dequantize_on_the_fly=False,
                 threads=1):
        """
        Initialize the PRS model.
        :param gdl: An instance of `GWADataLoader`.
        :param fix_params: A dictionary of fixed parameters.
        :param tracked_params: A list of parameters to track during the optimization.
        :param float_precision: The float precision to use for the model.
        :param lambda_min: The minimum value for the regularization parameter.
        :param verbose: Whether to print progress messages.
        :param low_memory: Whether to use low memory mode.
        :param dequantize_on_the_fly: Whether to dequantize the LD matrix on the fly.
        :param threads: The number of threads to use for parallel computation.
        """

        # ---------- General properties: ----------

        self.fix_params = fix_params or {}
        self.tracked_params = tracked_params or ['objective']
        self.float_precision = float_precision
        self.verbose = verbose
        self.gdl = gdl
        self.low_memory = low_memory
        self.threads = threads
        self.history = None

        # ----------------- Sanity checks -----------------

        assert gdl.ld is not None, "The LD matrices must be initialized in the GWADataLoader object."
        assert gdl.sumstats_table is not None, ("The summary statistics must be "
                                                "initialized in the GWADataLoader object.")

        # -------------------- The data --------------------

        self.chromosome = gdl.chromosomes[0]
        self.shape = self.gdl.shapes[self.chromosome]

        # >>>>>> Summary statistics <<<<<<
        ss_table = gdl.sumstats_table[self.chromosome]
        # Sample size per SNP:
        self.n_per_snp = ss_table.n_per_snp.astype(self.float_precision)
        self.std_beta = ss_table.get_snp_pseudo_corr().astype(self.float_precision)

        # Determine the overall sample size:
        self._sample_size = np.max(self.n_per_snp)

        # >>>>>> LD matrices <<<<<<

        ld_mat = self.gdl.ld[self.chromosome]

        # Determine how to load the LD data:
        if dequantize_on_the_fly and np.issubdtype(ld_mat.stored_dtype, np.integer):
            ld_dtype = ld_mat.stored_dtype
        else:
            if dequantize_on_the_fly:
                logging.warning("Dequantization on the fly is only supported for "
                                "integer data types. Ignoring this flag.")

            ld_dtype = float_precision
            dequantize_on_the_fly = False

        # self.ld_data, self.ld_indptr, self.ld_left_bound = ld_mat.load_data(
        #         return_symmetric=not low_memory,
        #         dtype=ld_dtype
        # )
        ld_lop = ld_mat.load(return_symmetric=not low_memory, dtype=ld_dtype)
        self.ld_data = ld_lop.ld_data
        self.ld_indptr = ld_lop.ld_indptr
        self.ld_left_bound = ld_lop.leftmost_idx

        # ------------- Data-dependent properties -------------

        self.dequantize_on_the_fly = dequantize_on_the_fly

        if self.dequantize_on_the_fly:
            info = np.iinfo(self.ld_data.dtype)
            self.dequantize_scale = 1. / info.max
        else:
            self.dequantize_scale = 1.
        
        if lambda_min is None or lambda_min == "infer":
            
            store_attrs = ld_mat.list_store_attributes()

            if 'Spectral properties' not in store_attrs or 'Extremal' not in ld_mat.get_store_attr('Spectral properties'):
                self.lambda_min = 0.1 #temporary fix
            else:
                self.lambda_min = ld_mat.get_lambda_min('min')
        else:
            self.lambda_min = lambda_min

        # ----------------- Model parameters -----------------

        # Coefficients:
        self.betas = None
        # Multiplication of coefficients with LD matrix:
        self.q = None
        # Coefficient differences (between iterations)
        self.betas_diff = None


    @property
    def m(self) -> int:
        """

        !!! seealso "See Also"
            * [n_snps][penprs.model.PRSModel.PRSModel.n_snps]

        :return: The number of variants in the model.
        """
        return self.gdl.m

    @property
    def n(self) -> int:
        """
        :return: The number of samples in the dataset.
        """
        return self._sample_size

    @property
    def n_snps(self) -> int:
        """
        !!! seealso "See Also"
            * [m][penprs.model.PRSModel.PRSModel.m]

        :return: The number of variants in the model.
        """
        return self.m

    @property
    def nnz(self) -> int:
        """
        :return: The number of non-zero coefficients.
        """
        assert self.betas is not None, "The coefficients BETA are not set. Call `.fit()` first."
        return np.count_nonzero(self.betas, axis=0)

    @property
    def max_betas_diff(self) -> float:
        """
        :return: The maximum absolute difference in the coefficients between iterations.
        """
        assert self.betas_diff is not None, "The coefficients BETA are not set. Call `.fit()` first."
        return np.max(np.abs(self.betas_diff))

    def _compute_lambda_max(self):
        """
        Compute the maximum LASSO penalty based on the input data.
        """
        return np.max(np.abs(self.n_per_snp * self.std_beta))

    def initialize(self, theta_0=None, param_0=None):

        self.initialize_hyperparameters(theta_0)
        self.initialize_model_parameters(param_0)
        self.init_history()

    def init_history(self):
        """
        Initialize the history dictionary.
        """

        self.history = {

        }

        for tt in self.tracked_params:
            if isinstance(tt, str):
                self.history[tt] = []
            elif callable(tt):
                self.history[tt.__name__] = []

        # Ensure that the objective is always tracked:
        if 'objective' not in self.history:
            self.history['objective'] = []

    def initialize_hyperparameters(self, theta_0=None):
        """
        Initialize the hyperparameters for the model.
        """
        raise NotImplementedError

    def initialize_model_parameters(self, param_0=None):

        param_0 = param_0 or {}

        self.q = np.zeros(self.n_snps, dtype=self.float_precision)

        if 'betas' in param_0:
            self.betas = param_0['betas'].astype(self.float_precision)
            self.p_gamma = np.count_nonzero(self.betas)
        else:
            self.betas = np.zeros(self.n_snps, dtype=self.float_precision)
            self.p_gamma = 0

        self.betas_diff = np.zeros(self.n_snps, dtype=self.float_precision)

    def get_coefficients(self):
        """
        Get the coefficients for the variants.
        :return: The coefficients for the variants.
        """
        return self.betas

    def fit(self, *args, **kwargs):
        """
        A genetic method to fit the PRS model.
        :raises NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError

    def objective(self):
        """
        :return: The objective function.
        """
        raise NotImplementedError

    def mse(self):
        """
        :return: The mean squared error.
        """

        return (1. - 2.*self.std_beta.dot(self.betas) + np.multiply(self.q, self.betas).sum(axis=0) +
            (1 + self.lambda_min)*(self.betas**2).sum(axis=0))

    def penalty(self):
        """
        :return: The penalty term.
        :raises NotImplementedError: If the method is not implemented in the child class.
        """

        raise NotImplementedError

    def predict(self, test_gdl=None):
        """
        Given the inferred effect sizes, predict the phenotype for the training samples in
        the GWADataLoader object or new test samples. If `test_gdl` is not provided, genotypes
        from training samples will be used (if available).

        :param test_gdl: A GWADataLoader object containing genotype data for new test samples.
        :raises ValueError: If the posterior means for BETA are not set. AssertionError if the GWADataLoader object
        does not contain genotype data.
        """

        if self.betas is None:
            raise ValueError("The coefficients BETA are not set. Call `.fit()` first.")

        if test_gdl is None:
            assert self.gdl.genotype is not None, "The GWADataLoader object must contain genotype data."
            test_gdl = self.gdl
            betas = self.betas
        else:
           betas = self.harmonize_data(gdl=test_gdl)

        return test_gdl.predict(betas)

    def harmonize_data(self, gdl=None, parameter_table=None):
        """
        Harmonize the inferred effect sizes with a new GWADataLoader object. This method is useful
        when the user wants to predict on new samples or when the effect sizes are inferred from a
        different set of samples. The method aligns the effect sizes with the SNP table in the
        GWADataLoader object.

        :param gdl: An instance of `GWADataLoader` object.
        :param parameter_table: A `pandas` DataFrame of variant effect sizes.

        :return: A tuple of the harmonized posterior inclusion probability, posterior mean for the effect sizes,
        and posterior variance for the effect sizes.

        """

        if gdl is None and parameter_table is None:
            return

        if gdl is None:
            gdl = self.gdl

        if parameter_table is None:
            parameter_table = self.to_table()
        else:
            parameter_table = parameter_table.loc[parameter_table['CHR'] == self.chromosome, ]

        snp_table = gdl.to_snp_table(col_subset=['SNP', 'A1', 'A2'],
                                     per_chromosome=True)[self.chromosome]

        from magenpy.utils.model_utils import merge_snp_tables

        beta_cols = expand_column_names('BETA', self.betas.shape)

        # Merge the effect table with the GDL SNP table:
        c_df = merge_snp_tables(snp_table, parameter_table, how='left',
                                signed_statistics=beta_cols)

        if len(c_df) < len(snp_table):
            raise ValueError("The parameter table could not aligned with the reference SNP table. "
                             "This may due to conflicts/errors in use of reference vs. "
                             "alternative alleles.")

        # Obtain the values for the coefficients:
        c_df[beta_cols] = c_df[beta_cols].fillna(0.)

        return c_df[beta_cols].values

    def to_table(self,  col_subset=('CHR', 'SNP', 'POS', 'A1', 'A2'), prune=False):
        """
        Output the model parameters as a `pandas` DataFrame.

        :param col_subset: A tuple of columns to include in the output table.
        :param prune: Whether to remove rows with zero effect sizes.

        :return: A `pandas` DataFrame containing the model parameters.
        """

        assert self.betas is not None, "The effect sizes are not set. Call `.fit()` first."

        if prune:
            # Check that there is only one set of effect sizes:
            assert self.betas.shape == (self.n_snps,)

        table = self.gdl.to_summary_statistics_table(col_subset=col_subset)

        from viprs.utils.compute_utils import expand_column_names

        table = pd.concat([
            table,
            pd.DataFrame(self.betas,
                         columns=expand_column_names('BETA', self.betas.shape),
                         index=table.index)
        ], axis=1)

        if prune:
            table = table.loc[table['BETA'] != 0, ].reset_index(drop=True)

        return table

    def to_history_table(self):
        """
        :return: A `pandas` DataFrame containing the history of tracked
        parameters as a function of the number of iterations.
        """
        return pd.DataFrame(self.history)

    def to_hyperparameter_table(self):
        """
        :return: A `pandas` DataFrame containing the hyperparameters of the model.
        """
        raise NotImplementedError

    def update_history(self):
        """
        Update the history dictionary with the current values of the
        tracked parameters and objectives.
        """

        for tt in self.tracked_params:

            if isinstance(tt, str) and hasattr(self, tt):
                attr = getattr(self, tt)
                if callable(attr):
                    self.history[tt].append(attr())
                else:
                    if tt=="betas":
                        self.history[tt].append(np.copy(attr))
                    else:
                        self.history[tt].append(attr)
            elif callable(tt):
                self.history[tt.__name__].append(tt(self))

    def pseudo_validate(self, test_gdl, metric='r2'):
        """
        Evaluate the prediction accuracy of the inferred PRS using external GWAS summary statistics.

        :param test_gdl: A `GWADataLoader` object with the external GWAS summary statistics and LD matrix information.
        :param metric: The metric to use for evaluation. Options: 'r2' or 'pearson_correlation'.

        :return: The pseudo-validation metric.
        """
        from penprs.eval.pseudo_metrics import pseudo_r2
        from viprs.eval.pseudo_metrics import pseudo_pearson_r

        metric = metric.lower()

        assert self.betas is not None, "The posterior means for BETA are not set. Call `.fit()` first."

        param_table = self.to_table()

        if metric in ('pearson_correlation', 'corr', 'r'):
            return pseudo_pearson_r(test_gdl, param_table)
        elif metric == 'r2':
            return pseudo_r2(test_gdl, param_table)
        elif metric == "pearson_r2":
            return pseudo_pearson_r(test_gdl, param_table)**2
        else:
            raise KeyError(f"Pseudo validation metric ({metric}) not recognized. "
                           f"Options are: 'r2' or 'pearson_correlation'.")

    def select_best_model(self, validation_gdl=None, criterion='objective'):

        """
        From the grid of models that were fit to the data, select the best
        model according to the specified `criterion`. If the criterion is the objective,
        the model with the highest objective will be selected. If the criterion is
        validation or pseudo-validation, the model with the highest R-squared on the
        validation set will be selected.

        :param validation_gdl: An instance of `GWADataLoader` containing data from the validation set.
        Must be provided if criterion is `validation` or `pseudo_validation`.
        :param criterion: The criterion for selecting the best model.
        Options are: (`objective`, `validation`, `pseudo_validation`)
        """

        # Sanity checks:
        assert criterion in ('objective', 'validation', 'pseudo_validation', 'pseudo_pearson_validation')
        assert self.betas is not None, "The effect sizes are not set. Call `.fit()` first."
        assert len(self.betas.shape) > 1, "Multiple models must be fit to select the best model."
        assert self.betas.shape[1] > 1, "Multiple models must be fit to select the best model."

        if criterion == 'objective':
            best_model_idx = np.argmax(self.objective())
        elif criterion == 'validation':

            assert validation_gdl is not None
            assert validation_gdl.sample_table is not None
            assert validation_gdl.sample_table.phenotype is not None

            from viprs.eval.continuous_metrics import r2

            # TODO: Fix the predict method
            prs = self.predict(test_gdl=validation_gdl)
            prs_r2 = np.array([r2(prs[:, i], validation_gdl.sample_table.phenotype)
                               for i in range(self.betas.shape[1])])
            best_model_idx = np.argmax(prs_r2)

        elif criterion == 'pseudo_validation':

            pseudo_r2 = self.pseudo_validate(validation_gdl, metric='r2')
            best_model_idx = np.argmax(np.nan_to_num(pseudo_r2, nan=0., neginf=0., posinf=0.))
        
        elif criterion == 'pseudo_pearson_validation':
            pseudo_r2 = self.pseudo_validate(validation_gdl, metric='pearson_r2')
            best_model_idx = np.argmax(np.nan_to_num(pseudo_r2, nan=0., neginf=0., posinf=0.))

        if int(self.verbose) > 1:
            logging.info(f"> Based on the {criterion} criterion, selected model: {best_model_idx}")
            logging.info("> Model details:\n")

        # -----------------------------------------------------------------------

        # Update the model parameters:
        self.betas = self.betas[:, best_model_idx]
        self.q = self.q[:, best_model_idx]

        # -----------------------------------------------------------------------

        return best_model_idx

    def set_model_parameters(self, parameter_table):
        """
        Parses a pandas dataframe with model parameters and assigns them
        to the corresponding class attributes.

        For example:
            * Columns with `BETA`, will be assigned to `self.betas`.

        :param parameter_table: A pandas dataframe.
        """

        self.betas = self.harmonize_data(parameter_table=parameter_table)

    def read_inferred_parameters(self, f_names, sep=r"\s+"):
        """
        Read a file with the inferred parameters.
        :param f_names: A path (or list of paths) to the file with the effect sizes.
        :param sep: The delimiter for the file(s).
        """

        if isinstance(f_names, str):
            f_names = [f_names]

        param_table = []

        for f_name in f_names:
            param_table.append(pd.read_csv(f_name, sep=sep))

        if len(param_table) > 0:
            param_table = pd.concat(param_table)
            self.set_model_parameters(param_table)
        else:
            raise FileNotFoundError

    def write_inferred_parameters(self, f_name, sep="\t"):
        """
        A convenience method to write the inferred posterior for the effect sizes to file.

        TODO:
            * Support outputting scoring files compatible with PGS catalog format:
            https://www.pgscatalog.org/downloads/#dl_scoring_files

        :param f_name: The filename where to write the effect sizes
        :param sep: The delimiter for the file (tab by default).
        """

        tables = self.to_table()

        if '.fit' not in f_name:
            ext = '.fit'
        else:
            ext = ''

        try:
            tables.to_csv(f_name + ext, sep=sep, index=False)
        except Exception as e:
            raise e



import numpy as np
import pandas as pd
from tqdm import tqdm
import logging

from .PRSModel import PRSModel
from ..utils.exceptions import OptimizationDivergence
from .co.coord_opt import cpp_update_delta, cpp_update_beta_ssl, cpp_update_var

class SSL(PRSModel):

    def __init__(self,
                 gdl,
                 l0=None,
                 l1=None,
                 hyperparam_update_freq=10,
                 unknown_var = False,
                 **prs_model_kwargs):

        super().__init__(gdl, **prs_model_kwargs)

        # Set SSL-specific hyperparameters:
        self.l0 = l0
        self.l1 = l1
        self.theta = None
        self.delta = None
        self.p_gamma = None  # Number of non-zero coefficients
        self.var = None
        self._a_theta_prior = None
        self._b_theta_prior = None
        self.hyperparam_update_freq = hyperparam_update_freq

        self.unknown_var = unknown_var
        self.prev_n_it = 0

    def initialize_hyperparameters(self, theta_0=None):

        if theta_0 is None:
            if self.fix_params is not None:
                theta_0 = self.fix_params
            else:
                theta_0 = {}
        else:
            if self.fix_params is not None:
                for key, value in self.fix_params.items():
                    theta_0.setdefault(key, value)

        if 'b' in theta_0:
            self._b_theta_prior = theta_0['b']
        elif self._b_theta_prior is None:
            self._b_theta_prior = self.n_snps

        if 'a' in theta_0:
            self._a_theta_prior = theta_0['a']
        elif self._a_theta_prior is None:
            # Set 'a' so that the prior mean of theta is 0.05:
            self._a_theta_prior = (0.05/(1. - 0.05))*self._b_theta_prior

        # Default settings for the SSL hyperparameters:
        lam_max = self._compute_lambda_max()

        if 'l1' in theta_0:
            self.l1 = theta_0['l1']
        elif 'min_lambda_frac' in theta_0:
            self.l1 = theta_0['min_lambda_frac']*lam_max
        elif self.l1 is None:
            # By default, scale lambda_max by 10^-3:
            self.l1 = 1e-3*lam_max

        if 'l0' in theta_0:
            self.l0 = theta_0['l0']
        elif 'max_lambda_frac' in theta_0:
            self.l0 = theta_0['max_lambda_frac']*lam_max
        elif self.l0 is None:
            # By default set lambda_0 to 10% of lambda_max
            # or 100*l1:
            self.l0 = max(0.1*lam_max, 100*self.l1)

        if self.unknown_var:
            self.min_var, self.init_var = self._init_min_var(self.n)
        else:
            self.min_var, self.init_var = 0., 0.


        if self.unknown_var:
            self.var = self.init_var
        elif 'var' in theta_0:
            self.var = theta_0['var']
        else:
            self.var = 1.

        if 'theta' in theta_0:
            self.theta = theta_0['theta']
        else:
            # Set theta based on the mean of the Beta prior:
            self.theta = self._a_theta_prior / (self._a_theta_prior + self._b_theta_prior)


        self._a_theta_prior = np.dtype(self.float_precision).type(self._a_theta_prior)
        self._b_theta_prior = np.dtype(self.float_precision).type(self._b_theta_prior)
        self.l0 = np.dtype(self.float_precision).type(self.l0)
        self.l1 = np.dtype(self.float_precision).type(self.l1)
        self.var = np.dtype(self.float_precision).type(self.var)
        self.lambda_min = np.dtype(self.float_precision).type(self.lambda_min)
        self.theta = np.dtype(self.float_precision).type(self.theta)

        self._update_delta()

    def _init_min_var(self, n):
        """
        Initialize the minimum variance for the unknown variance case.
        """

        from scipy.stats import chi2

        df = 3
        sigquant = 0.9
        sigest=1
        qchi = chi2.ppf(1 - sigquant, df)
        ncp = (sigest ** 2) * qchi / df
        min_var = (sigest ** 2) / n
        init_var = df * ncp / (df + 2)

        return min_var, init_var

    def _p_star(self):
        """
        Compute the p-star quantity for the spike-and-slab prior.
        """
        return 1./(1. + ((1. - self.theta)/self.theta)*(self.l0/self.l1) *
                   np.exp(-np.abs(self.betas) * (self.l0-self.l1)))

    def _lambda_star(self):
        """
        Compute the lambda-star quantity for the spike-and-slab prior.
        """
        p_star = self._p_star()
        return self.l1*p_star + self.l0*(1. - p_star)

    def _update_theta(self):
        """
        Update the theta parameter based on the current value of the coefficients
        and hyperparameters.
        """

        self.theta = (self._a_theta_prior + self.nnz) / (
            self._a_theta_prior + self._b_theta_prior + self.n_snps
        )

    def _update_delta(self):
        """
        Update the hard-thresholding delta parameter based
        on the current value of the coefficients and hyperparameters.
        """

        # NOTE: Due to a bug in cython dynamic type
        # inference, we need to cast the variables to float64
        # here and then cast the result back to the desired precision
        self.delta = np.dtype(self.float_precision).type(
            cpp_update_delta(
                np.dtype('float64').type(self.n),
                np.dtype('float64').type(self.l0),
                np.dtype('float64').type(self.l1),
                np.dtype('float64').type(self.theta),
                np.dtype('float64').type(self.var)
            )
        )

    def _coord_ascent_step(self, update_var=False):
        """
        Perform a single iteration of the coordinate ascent algorithm
        for the SSL model, given the current settings of the
        coefficients and hyperparameters.
        """

        # Wrap the hyperparameters in numpy arrays to pass by reference
        # to cython/c++:
        delta = np.array([self.delta],dtype=self.float_precision)
        theta = np.array([self.theta],dtype=self.float_precision)
        var = np.array([self.var],dtype=self.float_precision)
        p_gamma = np.array([self.p_gamma], dtype=np.int32)

        cpp_update_beta_ssl(
            self.ld_left_bound, self.ld_indptr, self.ld_data,
            self.std_beta,
            self.betas, self.betas_diff,
            self.q, self.n_per_snp,
            p_gamma, theta, delta, var,
            self.n, self.l0, self.l1,
            self._a_theta_prior, self._b_theta_prior,
            self.hyperparam_update_freq,
            self.lambda_min,
            self.init_var, self.min_var, update_var,
            self.dequantize_scale,
            self.threads,
            self.low_memory
        )

        # Retrieve updated values
        self.delta = np.dtype(self.float_precision).type(delta[0])
        self.theta = np.dtype(self.float_precision).type(theta[0])
        self.var = np.dtype(self.float_precision).type(var[0])
        self.p_gamma = p_gamma[0]

    def _update_var(self):
        # self.var = self.n/(self.n+2) * self.mse()
        self.var = np.dtype(self.float_precision).type(
            cpp_update_var(
                np.dtype('float64').type(self.n),
                np.dtype('float64').type(self.betas),
                np.dtype('float64').type(self.std_beta),
                np.dtype('float64').type(self.q),
                np.dtype('float64').type(self.lambda_min)
            )
        )


    def penalty(self):
        """
        Compute the contribution of the penalty for the SSL objective
        using the current value of the coefficients and hyperparameters.
        """

        p_star_0 = -self.n_snps*np.log(1. + ((1. - self.theta)/self.theta)*(self.l0/self.l1))

        return -self.l1*np.abs(self.betas).sum() + p_star_0 - np.log(self._p_star()).sum(axis=0)

    def log_likelihood(self):
        """
        Compute the log-likelihood of the SSL model given the current
        value of the coefficients and hyperparameters.
        """
        return -self.n/(2*self.var) * self.mse() - (self.n+2) * np.log(np.sqrt(self.var))

    def objective(self):
        """
        The objective function for the SSL model.
        """
        return self.log_likelihood() + self.penalty()

    def fit(self,
            max_iter=1000,
            theta_0=None,
            param_0=None,
            continued=False,
            min_iter=3,
            f_abs_tol=1e-4,
            x_abs_tol=1e-4,
            update_var=False,
            disable_pbar=False):

        # initialize parameters
        if not continued:
            self.initialize(theta_0, param_0)
            self.update_history()

        pbar = tqdm(range(max_iter),
                    disable=not self.verbose or disable_pbar,
                    desc=f'Chromosome {self.chromosome} ({self.n_snps} variants)')

        curr_obj = prev_obj = self.history['objective'][-1]
        stop_iter = False

        for i in pbar:

            if stop_iter:
                pbar.set_postfix({
                    'Final objective': f"{curr_obj:.3f}",
                    'Nonzero': self.p_gamma
                })
                pbar.n = i - 1
                pbar.total = i - 1
                pbar.refresh()
                pbar.close()
                
                self.prev_n_it = i-1
                break

            # Run the coordinate ascent step:
            self._coord_ascent_step(update_var=update_var)
            # Update the tracked parameters (including objectives):
            self.update_history()

            # Extract the current value for the objective:
            curr_obj = self.history['objective'][-1]

            # Update the progress bar:
            pbar.set_postfix(
                {
                    'Objective': f"{curr_obj:.3f}",
                    'Nonzero': self.p_gamma
                }
            )

            if i > min_iter:
                # Check for convergence:
                if not np.isfinite(curr_obj):
                    raise OptimizationDivergence(f"Stopping at iteration {i}: "
                                                 "The optimization algorithm is not converging!\n"
                                                 f"The objective is undefined ({curr_obj}).")
                # elif self.mse() < 0:
                #     raise OptimizationDivergence(f"Stopping at iteration {i}: "
                #                                  "The optimization algorithm is not converging!\n"
                #                                  f"The MSE is negative ({self.mse()}).")
                elif np.isclose(prev_obj, curr_obj, atol=f_abs_tol, rtol=0.):
                    stop_iter = True
                elif self.max_betas_diff < x_abs_tol:
                    stop_iter = True

        return self

    def warm_start_fit(self,
                        l0_ladder=None,
                        ladder_steps=20,
                        max_lambda_frac=.99,
                        pathwise_fit = False,
                        pathwise_decrease_l1 = False,
                        save_intermediate=False,
                        theta_0=None,
                        param_0=None,
                        **fit_kwargs):

        """
        Fit the model using a warm-start strategy, where the model is fit
        multiple times with different values of the l0 hyperparameter.

        :param l0_ladder: A list of values for the l0 hyperparameter to use
            in the warm-start strategy. If None, a default ladder of values
            will be used.
        :param ladder_steps: The number of steps to use in the ladder of l0 values.
        :param max_lambda_frac: The multiplier to use to compute the maximum
            lambda value for the default ladder of l0 values.
        :param save_intermediate: If True, the model parameters will be saved
            for each `l0` value.
        :param theta_0: Initial values for the hyperparameters.
        :param param_0: Initial values for the model parameters.
        :param fit_kwargs: Additional keyword arguments to pass to the `fit`
        """

        assert l0_ladder is not None or ladder_steps is not None, (
            "Either `l0_ladder` or `ladder_steps` must be provided."
        )

        if self.betas is None:
            self.initialize(theta_0=theta_0, param_0=param_0)

        if l0_ladder is None:

            # Generate a default ladder of l0 values by
            # interpolating from `l1` to lam_max on a log2 scale:
            max_penalty = max_lambda_frac*self._compute_lambda_max()


            # run warm-start fit in a pathwise style or default 
            if pathwise_fit:

                l0_ladder = np.logspace(np.log2(self.l1),
                np.log2(max_penalty),
                num=ladder_steps,
                base=2)

                l0_ladder = np.sort(l0_ladder)
                l0_ladder = l0_ladder[::-1]

                if pathwise_decrease_l1:
                    self.l0 = 0.99*max_penalty
            else:

                l0_ladder = np.logspace(
                    np.log2(self.l1),
                    np.log2(max_penalty),
                    num=ladder_steps,
                    base=2)

        if save_intermediate:
            betas = np.empty((self.n_snps, len(l0_ladder)), dtype=self.float_precision)
            qs = np.empty((self.n_snps, len(l0_ladder)), dtype=self.float_precision)

        pbar = tqdm(range(len(l0_ladder)),
                    total=len(l0_ladder),
                    disable=not self.verbose,
                    desc=f'Chromosome {self.chromosome} ({self.n_snps} variants)')

        for i in pbar:
            
            if pathwise_decrease_l1:
                self.l1 = l0_ladder[i]
            else:
                self.l0 = l0_ladder[i]

            if i == 0:
                self.update_history()

            update_var = False

            if i !=0 and self.unknown_var:
                if self.prev_n_it < 100 :
                    update_var = True
                    self._update_var()

                    if self.var < self.min_var:
                        self.var = self.init_var
                        update_var = False
                else:
                    update_var = False
                    if self.prev_n_it == 1000:
                        self.var = self.init_var

            self.fit(continued=True, disable_pbar=True, update_var=update_var, **fit_kwargs)

            pbar.set_postfix(
                {
                    'Objective': f"{self.objective():.3f}",
                    'Nonzero': self.p_gamma,
                    'l0': f"{self.l0:.2f}",
                    'var': f"{self.var:.3f}"
                }
            )

            if save_intermediate:
                betas[:, i] = self.betas
                qs[:, i] = self.q

        if save_intermediate:
            self.betas = betas
            self.q = qs
            self.l0 = l0_ladder

        return self

    def select_best_model(self, validation_gdl=None, criterion='objective'):

        # Call the parent method:
        best_idx = super().select_best_model(validation_gdl=validation_gdl, criterion=criterion)

        # Update the hyperparameter:
        self.l0 = self.l0[best_idx]

        # Update the rest of the hyperparameters:
        self._update_theta()
        self._update_delta()

    def to_hyperparameter_table(self):

        return pd.DataFrame([
            {'Parameter': 'delta', 'Value': self.delta},
            {'Parameter': 'theta', 'Value': self.theta},
            {'Parameter': 'var', 'Value': self.var},
            {'Parameter': 'l0', 'Value': self.l0},
            {'Parameter': 'l1', 'Value': self.l1},
            {'Parameter': 'Nonzero coefficients', 'Value': self.nnz},
            {'Parameter': 'lambda_min', 'Value': self.lambda_min},
            {'Parameter': 'a (theta prior)', 'Value': self._a_theta_prior},
            {'Parameter': 'b (theta prior)', 'Value': self._b_theta_prior},
        ])



import numpy as np
import magenpy as mgp
from penprs.model.Lasso import Lasso
from penprs.model.SSL import SSL
from viprs.model.vi.e_step_cpp import check_blas_support, check_omp_support
from functools import partial
import shutil
import pytest


@pytest.fixture(scope='module')
def gdl_object():
    """
    Initialize a GWADataLoader using data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    gdl = mgp.GWADataLoader(mgp.tgp_eur_data_path(),
                            sumstats_files=mgp.ukb_height_sumstats_path(),
                            sumstats_format='fastgwa',
                            backend='xarray')

    ld_block_url = ("https://bitbucket.org/nygcresearch/ldetect-data/raw/"
                    "ac125e47bf7ff3e90be31f278a7b6a61daaba0dc/EUR/fourier_ls-all.bed")
    gdl.compute_ld('block', gdl.output_dir, ld_blocks_file=ld_block_url)

    gdl.harmonize_data()

    yield gdl

    # Clean up after tests are done:
    gdl.cleanup()
    shutil.rmtree(gdl.temp_dir)
    shutil.rmtree(gdl.output_dir)


@pytest.fixture(scope='module')
def lasso_model(gdl_object):
    """
    Initialize a basic Lasso model using GWAS sumstats data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    return Lasso(gdl_object)


@pytest.fixture(scope='module')
def ssl_model(gdl_object):
    """
    Initialize a basic SSL model using GWAS sumstats data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    return SSL(gdl_object)

class TestSSL(object):

    def test_init(self,
                  ssl_model: SSL,
                  gdl_object: mgp.GWADataLoader):

        assert ssl_model.m == gdl_object.m

        ssl_model.initialize()

        # Check the input data:
        for p in (ssl_model.std_beta, ssl_model.n_per_snp):
            assert p.shape == (ssl_model.m, )

        # Check the LD data:
        assert ssl_model.ld_indptr.shape == (ssl_model.m + 1, )
        assert ssl_model.ld_left_bound.shape == (ssl_model.m, )
        assert ssl_model.ld_data.shape == (ssl_model.ld_indptr[-1], )

        # Check hyperparameters:
        assert np.all(np.asarray(ssl_model.l0) >= ssl_model.l1)
        assert 0. < ssl_model.var
        assert 0 <= ssl_model.theta <=1
        assert ssl_model._a_theta_prior > 0
        assert ssl_model._b_theta_prior > 0

    def test_fit(self, ssl_model: SSL):

        ssl_model.fit(max_iter=10)

        assert ssl_model.betas.shape == (ssl_model.m,)

        # Test that the following methods are working properly:
        ssl_model.to_table()
        ssl_model.to_history_table()
        ssl_model.mse()
        ssl_model.log_likelihood()
        ssl_model.penalty()
        ssl_model.objective()

class TestLasso(object):

    def test_init(self,
                  lasso_model: Lasso,
                  gdl_object: mgp.GWADataLoader):
        """
        Verify initialization and hyperparameter setup for the Lasso model.
        """

        # Check that dimensions match
        assert lasso_model.m == gdl_object.m

        # Initialize model
        lasso_model.initialize()

        # Check input data arrays
        for p in (lasso_model.std_beta, lasso_model.n_per_snp):
            assert p.shape == (lasso_model.m, )

        # Check LD-related arrays
        assert lasso_model.ld_indptr.shape == (lasso_model.m + 1, )
        assert lasso_model.ld_left_bound.shape == (lasso_model.m, )
        assert lasso_model.ld_data.shape == (lasso_model.ld_indptr[-1], )

        # Hyperparameter checks
        assert lasso_model.lam is not None
        assert lasso_model.lam >= 0
        assert lasso_model.lambda_min >= 0

    def test_fit(self, lasso_model: Lasso):
        """
        Test that Lasso fitting, objective computation, and reporting methods work.
        """

        # Run a short fit to ensure flow executes
        lasso_model.fit(max_iter=10)

        # Betas should be correctly shaped
        assert lasso_model.betas.shape == (lasso_model.m,)

        # Test that main methods execute successfully
        lasso_model.to_hyperparameter_table()
        lasso_model.to_history_table()
        lasso_model.mse()
        lasso_model.penalty()
        lasso_model.objective()

        # Check objective consistency
        lasso_model.penalty()
        lasso_model.mse()
        lasso_model.objective()

@pytest.mark.xfail(not check_blas_support(), reason="BLAS library not found!")
def test_check_blas_support():
    assert check_blas_support()


@pytest.mark.xfail(not check_omp_support(), reason="OpenMP library not found!")
def test_check_omp_support():
    assert check_omp_support()



import numpy as np
import pandas as pd
import itertools


class HyperparameterGrid(object):

    def __init__(self,
                l0_grid=None,
                l0_steps=None,
                l1_grid=None,
                l1_steps=None,
                var_grid = None,
                var_steps = None,
                a_grid = None,
                a_steps = None,
                lambda_max= None,
                scale_l0_by_l1=False,
                n_snps=12500):

        self.n_snps = n_snps
        self.lambda_max = lambda_max
        self.scale_l0_by_l1 = scale_l0_by_l1
        self._search_params = []

        # Initialize the grid for l0:
        self.l0 = l0_grid
        if self.l0 is not None:
            self._search_params.append('l0')

            #ensure correctly parsed into list of values
            if isinstance(self.l0, str):
                self.l0 = np.array([np.float64(l0.strip()) for l0 in self.l0.split(",")])
        elif l0_steps is not None:
            self.generate_l0_grid(steps=l0_steps)


        # Initialize the grid for l1:
        self.l1 = l1_grid
        if self.l1 is not None:
            self._search_params.append('l1')

            #ensure correctly parsed into list of values
            if isinstance(self.l1, str):
                self.l1 = np.array([np.float64(l1.strip()) for l1 in self.l1.split(",")])
        elif l1_steps is not None:
            self.generate_l1_grid(steps=l1_steps)

        # Initialize the grid for var:
        self.var = var_grid
        if self.var is not None:
            self._search_params.append('var')

            #ensure correctly parsed into list of values
            if isinstance(self.var, str):
                self.var = np.array([np.float64(var.strip()) for var in self.var.split(",")])
        elif var_steps is not None:
            self.generate_var_grid(steps=var_steps)

        # Initialize the grid for a:
        self.a = a_grid
        if self.a is not None:
            self._search_params.append('a')

            #ensure correctly parsed into list of values
            if isinstance(self.a, str):
                self.a = np.array([np.int32(a.strip()) for a in self.a.split(",")])
        elif a_steps is not None:
            self.generate_a_grid(steps=a_steps)


    def generate_l0_grid(self, steps=50):
        """
        Generate a grid of values for the `l0` spike penalty hyperparameter.

        :param steps: The number of steps for the l0 grid.
        """

        assert steps > 0
        if self.lambda_max is not None:
            if self.scale_l0_by_l1:
                self.l0 = np.logspace(0, 2, steps)
            else:
                self.l0 = np.linspace(0.01,1,steps=steps) * self.lambda_max
        else:

            start = 200
            end = 2000
            
            self.l0 = np.linspace(start, end, steps)

        if 'l0' not in self._search_params:
            self._search_params.append('l0')
    
    def generate_l0_per_of_lm_grid(self):
        self.l0 = self.l0 * self.lambda_max

    def generate_l1_grid(self, steps=3):
        """
        Generate a grid of values for the `l1` slab penalty hyperparameter.

        :param steps: The number of steps for the l1 grid.
        """

        assert steps > 0

        # print(self.lambda_max)
        if self.lambda_max is not None:
            eps = 0.001
            self.l1 = np.logspace(np.log10(self.lambda_max), np.log10(eps * self.lambda_max), steps)
        else:
            half_range = (steps - 1) // 2
            if steps % 2 == 1:
                exponents = np.linspace(-half_range, half_range, steps)
            else:
                exponents = np.concatenate([np.linspace(-half_range-1, 0, steps//2+1), np.linspace(1, half_range, steps//2-1)])

            self.l1 = 10 ** exponents 

        if 'l1' not in self._search_params:
            self._search_params.append('l1')
    
    def generate_var_grid(self, steps=3):
        """
        Generate a grid of values for the `var` or sigma2 error variance.

        :param steps: The number of steps for the var grid.
        """

        assert steps > 0

        start = 0.6
        end = 1.1

        self.var = np.concatenate([np.linspace(start, 1, steps//2), np.linspace(1, end, steps//2+1)])

        if 'var' not in self._search_params:
            self._search_params.append('var')

    def generate_a_grid(self, steps=6):
        """
        Generate a grid of values for the `l1` slab penalty hyperparameter.

        :param steps: The number of steps for the l1 grid.
        """

        assert steps > 0

        self.a = np.linspace(1, 0.5*self.n_snps, steps)

        if 'a' not in self._search_params:
            self._search_params.append('a')

    def combine_grids(self):
        """
        Weave together the different hyperparameter grids and return a list of
        dictionaries where the key is the hyperparameter name and the value is
        value for that hyperparameter.

        :return: A list of dictionaries containing the hyperparameter values.
        :raises ValueError: If all the grids are empty.

        """
        hyp_names = [name for name, value in self.__dict__.items()
                     if value is not None and name in self._search_params]

        if len(hyp_names) > 0:
            hyp_values = itertools.product(*[hyp_grid for hyp_name, hyp_grid in self.__dict__.items()
                                             if hyp_grid is not None and hyp_name in hyp_names])

            return [dict(zip(hyp_names, hyp_v)) for hyp_v in hyp_values]
        else:
            raise ValueError("All the grids are empty!")

    def to_table(self):
        """
        :return: The hyperparameter grid as a pandas `DataFrame`.
        """

        combined_grids = self.combine_grids()
        if combined_grids:
            return pd.DataFrame(combined_grids)
